import fs from 'node:fs';
import path from 'node:path';
import { defineCommand } from 'citty';
import { Client } from 'ssh2';
import { relative, resolve } from 'pathe';

async function createResolver() {
  const jiti = await import('jiti').then((r) => r.default || r);
  const _jitiRequire = jiti(process.cwd(), {
    cache: true,
    esmResolve: true,
    requireCache: false,
    interopDefault: true
  });
  const _import = (id) => {
    const r = _jitiRequire(id);
    return Promise.resolve(r.default || r);
  };
  const resolve = (id) => _jitiRequire.resolve(id);
  const tryResolve = (id) => {
    try {
      return resolve(id);
    } catch {
    }
  };
  return {
    relative: (path) => relative(process.cwd(), path),
    formateRelative: (path) => `\`./${relative(process.cwd(), path)}\``,
    import: _import,
    resolve,
    tryResolve
  };
}

const sftpReadDir = (jumpClient, path) => {
  return new Promise((resolve) => {
    jumpClient.sftp((sftpErr, sftp) => {
      if (sftpErr) {
        return resolve({ error: sftpErr, list: void 0 });
      }
      sftp.readdir(path, (err, list) => {
        if (err) {
          return resolve({ error: err, list: void 0 });
        }
        return resolve({ error: err, list });
      });
    });
  });
};

const deploy = defineCommand({
  args: {
    environment: {
      type: "positional",
      required: true,
      description: "Environment"
    },
    workDir: {
      type: "string",
      required: false,
      description: "working directory"
    }
  },
  async run({ args }) {
    const { workDir, environment } = args;
    const resolver = await createResolver();
    const client = new Client();
    const configPath = resolve(workDir, `simple-deploy.config.ts`);
    if (!fs.existsSync(configPath)) {
      console.log("Can't find config file `simple-deploy.config.ts`");
      return;
    }
    const config = await resolver.import(configPath);
    const deployConfig = config?.deploy?.[environment];
    if (deployConfig) {
      if (deployConfig.proxyJump) {
        const jumpConfig = config?.deploy?.[deployConfig.proxyJump];
        if (!jumpConfig) {
          console.log("proxyJump is not defined");
          return;
        }
        const jumpClient = new Client();
        console.log(`Connecting to server ${jumpConfig.host}...`);
        client.on("ready", () => {
          client.forwardOut(
            "127.0.0.1",
            8124,
            deployConfig.host,
            22,
            (err, stream) => {
              if (err) {
                throw err;
              }
              console.log(`Jumping to server ${deployConfig.host}...`);
              jumpClient.connect({
                sock: stream,
                host: deployConfig.host,
                privateKey: deployConfig.privateKey,
                username: deployConfig.user
              });
            }
          );
        }).connect({
          username: jumpConfig.user,
          host: jumpConfig.host,
          privateKey: jumpConfig.privateKey,
          agentForward: true,
          agent: process.env.SSH_AUTH_SOCK
        });
        jumpClient.on("ready", async () => {
          let deployCommand = [
            `git clone --single-branch -b main --progress --verbose ${deployConfig.repo} ${deployConfig.path}`,
            `cd ${deployConfig.path}`
          ];
          const postDeployCmd = deployConfig.postDeploy || "";
          const gitConfigPath = path.join(deployConfig.path, ".git");
          const { error, list } = await sftpReadDir(jumpClient, gitConfigPath);
          if (error) {
            throw error;
          }
          if (list) {
            deployCommand = [
              `cd ${deployConfig.path}`,
              "git clean -d -f",
              "git fetch origin main:origin/main",
              "git reset --hard origin/main"
            ];
          }
          if (postDeployCmd) {
            deployCommand.push(postDeployCmd);
          }
          jumpClient.exec(deployCommand.join("\n"), (err, stream) => {
            if (err) {
              throw err;
            }
            stream.on("data", (data) => {
              console.log(data.toString());
            });
            stream.stderr.on("data", function(data) {
              const output = data.toString();
              if (output.startsWith("fatal")) {
                console.log(output);
              } else {
                console.log(output);
              }
            });
            stream.on("close", () => {
              jumpClient.destroy();
              client.destroy();
            });
          });
          console.log("jumpClient already!");
        });
      }
    } else {
      console.log("Config not found");
    }
  }
});

export { deploy as default };
